import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.zip.GZIPInputStream;

import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;

import edu.stanford.nlp.ie.AbstractSequenceClassifier;
import edu.stanford.nlp.ie.crf.CRFClassifier;
import edu.stanford.nlp.ling.CoreAnnotations;
import edu.stanford.nlp.ling.CoreLabel;
import edu.stanford.nlp.ling.HasWord;
import edu.stanford.nlp.ling.Word;
import edu.stanford.nlp.ling.tokensregex.parser.ParseException;
import edu.stanford.nlp.process.DocumentPreprocessor;


public class FindPassages {
	
	//args
	public static final String GIGA_DIR_TAG = "-gigaDir";
	public static final String KEY_FILE_TAG = "-keyFile";
	public static final String OUT_FILE_TAG = "-outFile";
	public static final String PARSE_TAG = "-parse";
	public static final String NER_TAG = "-ner";
	public static final String THREAD_TAG = "-thread";
	
	
	//xml tags for WDW
	static final public String XML_HEAD_TAG="<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
	static final public String ROOT_TAG="ROOT";
	static final public String MC_TAG="mc";

	
	
	public static void main(String[] args) throws IOException, ClassCastException, ClassNotFoundException, InterruptedException {
		String gigaDir = null;
		String keyFile = null;
		String outFile = null;
		boolean isParse = false;
		String nerPath = "";
		int nThreads = 1;
		
		//read arguments
		for(int i=0;i<args.length;i++){
			String a = args[i];
			if(a.equals(GIGA_DIR_TAG)){
				gigaDir = args[++i];
				System.out.println(String.format("%s=%s", GIGA_DIR_TAG, gigaDir));
			}else if(a.equals(KEY_FILE_TAG)){
				keyFile = args[++i];
				System.out.println(String.format("%s=%s", KEY_FILE_TAG, keyFile));
			}else if(a.equals(OUT_FILE_TAG)){
				outFile = args[++i];
				System.out.println(String.format("%s=%s", OUT_FILE_TAG, outFile));
			}else if(a.equals(PARSE_TAG)){
				isParse = true;
				System.out.println(String.format("%s=%s", PARSE_TAG, (isParse)? "true" : "false"));
			}else if(a.equals(NER_TAG)){
				nerPath = args[++i];
			}else if(a.equals(THREAD_TAG)){
				nThreads = Integer.parseInt(args[++i]);
				System.out.println(String.format("%s=%d", THREAD_TAG, nThreads));
			}else{
				System.err.println("invalid argment : "+a);
				return;
			}
		}
		
		//check arguments
		if(gigaDir==null || keyFile==null || outFile==null){
			System.err.println(String.format("invalid option : %s, %s and %s are required", GIGA_DIR_TAG, KEY_FILE_TAG, OUT_FILE_TAG ));
			return;
		}
		if(isParse==true && nerPath.isEmpty()){
			System.err.println("NER is required to parse :"+NER_TAG );
			return;
		}
		
		
		System.out.println("mapping gigaword files");
		//super ids -> file name
		Map<String, String> superId2File = new HashMap<String, String>();
		List<String> pathes = getFiles(gigaDir, ".gz");
		for(String f : pathes) superId2File.put((new File(f)).getName().toLowerCase().replace(".gz", ""), f);
		
		//read keyFiles
		Document key= Jsoup.parse(new File(keyFile), "UTF-8");
		Elements mcqs = key.getElementsByTag(MC_TAG); 
		
		//sort question by super id of its passage
		Map<String, Set<Element>> superId2Q = new HashMap<String, Set<Element>>();
		System.out.println("sorting keys...");
		for(Element mcq : mcqs){
			try {
				for(Element doc : mcq.getElementsByTag(SuperID_Process.CONTEXT_ART_TAG)){
					String id =  doc.attr(SuperID_Process.ID_TAG);
					//check
					String superId =id.split("\\.")[0];
					superId = superId.substring(0, superId.length()-2).toLowerCase(); //cut last 2 ch which is day
					if(!superId2File.containsKey(superId)) throw new Exception("no super id in Gigaword: "+ superId);
					
					if(!superId2Q.containsKey(superId)) superId2Q.put(superId, new HashSet<Element>());
					superId2Q.get(superId).add(mcq);
				}				
			} catch (Exception e) {
				e.printStackTrace();
			}
			//if(i%Math.max(1, keys.size()/20)==0) System.out.print("#");
		}
		//System.out.println("]");
		System.out.println("reading "+superId2Q.size()+"-gigaword files");
		
		//NER for each processing
		AbstractSequenceClassifier<CoreLabel> ner = null;
		if(nerPath!=""){
			System.err.println("Gigaword will be parsed.");
			nThreads=1;
			ner = CRFClassifier.getClassifier(nerPath);
		}
		
		//read file by superid
		SuperID_Process[] superIdPros = new SuperID_Process[nThreads];
		
		//processing for each super id
		for(String sId : superId2Q.keySet()){
			int idx = getSlot(superIdPros);
			superIdPros[idx] = new SuperID_Process(superId2Q, superId2File, isParse, ner, sId);
			superIdPros[idx].start();
			
			System.out.println("passage in : "+sId);
		}
		//wait to join
		for(int i=0;i<superIdPros.length;i++) superIdPros[i].join();
		
		
		//write xmlMCquestion
		File outXml = new File(outFile);
		BufferedWriter bf = new BufferedWriter(new FileWriter( outXml ));
		bf.write(XML_HEAD_TAG); bf.newLine();
		bf.write("<"+ROOT_TAG+">"); bf.newLine();
		for(Element mcq : key.getElementsByTag(MC_TAG)){
			bf.write(mcq.toString());
			bf.newLine();
		}
		bf.write("</"+ROOT_TAG+">");
		bf.close();
//		Set<XmlMCQuestion> mcqs = new HashSet<XmlMCQuestion>();
//		for(Set<XmlMCQuestion> s : superId2Q.values()) mcqs.addAll(s);
//		int c=0;
//		for(XmlMCQuestion mcq : mcqs){
//			try {
//				mcq.toFile(outDir);
//				c++;
//			} catch (IOException e) {
//				e.printStackTrace();
//			}
//		}
		
//		System.out.println(countS+"-questions are regenerated");
		
		//check
//		Document result= Jsoup.parse(new File(opt.outFile), "UTF-8");
//		for(Element mcq : result.getElementsByTag(MC_TAG))for(Element doc : mcq.getElementsByTag(CONTEXT_ART_TAG)){
//			if(!doc.hasText()) System.out.println("no passage found : "+mcq.attr(ID_TAG));
//		}
		System.out.println("DONE");
	}
	
	private static int getSlot(Thread[] ths){
		int idx = -1;
		int i = 0;
		while(idx<0){
			if(ths[i]==null || !ths[i].isAlive())idx=i;
			i=++i%ths.length;
		}
		return idx;
	}
	
	//collect all gigaword files
//	private static List<String> getGigaFiles(String dir){
//		
//		File d = new File(dir);
//	    File[] files = d.listFiles();
//	    List<String> path = new ArrayList<String>();
//	    for(File f : files){
//	    	if(f.isDirectory()){
//	    		path.addAll(getGigaFiles(f.getAbsolutePath()));
//	    	}else{
//	    		if(!f.getName().contains(".") && dir.contains("/data")) path.add(f.getAbsolutePath());
//	    	}
//	    }
//	    return path;
//	}
	
	//collect key files
	private static List<String> getFiles(String dir, String type){
		
		File d = new File(dir);
	    File[] files = d.listFiles();
	    List<String> path = new ArrayList<String>();
	    
	    for(File f : files){
	    	if(f.isDirectory()){
	    		path.addAll(getFiles(f.getAbsolutePath(), type));
	    	}else{
	    		if(f.getName().endsWith(type)) path.add(f.getAbsolutePath());
	    	}
	    }
	    return path;
	}
}

class SuperID_Process extends Thread{
	
	//xml tags
	static final public String INDEX_TAG="idx";
	static final public String CAND_TAG="choice";
	static final public String LEFT_CONTEXT_TAG="leftcontext";
	static final public String RIGHT_CONTEXT_TAG="rightcontext";
	static final public String LEFT_BLANK_TAG="leftblank";
	static final public String RIGHT_BLANK_TAG="rightblank";
	static final public String BLANK_TAG="blank";
	static final public String QUESTION_TAG="question";
	static final public String ID_TAG="id";
	static final public String CONTEXT_ART_TAG="contextArt";
	static final public String ANS_TYPE="PERSON";
	
	static final public String ID_TAIL ="</DOC>";
	static final public String TEXT =  "<TEXT>";
	static final public String TEXT_TAIL = "</TEXT>";
	static final public String CORRUPTION = "<corruption>";
	static final public String CORRUPTION_TAIL = "</corruption>";
	static final public String PARAGRAPH  = "<P>";
	static final public String PARAGRAPH_TAIL = "</P>";
	
	//tagging marker
	static final public String ENTITY_TAG_HEAD="@ENTITY:";
	static final public String ENTITY_TAG_TAIL=":ENTITY@";
	
	//processing tools
	private AbstractSequenceClassifier<CoreLabel> ner = null;
	private String sId="UNKNOWN";
	
	
	static Map<String, Set<Element> > superId2Q=null;
	static Map<String, String> superId2File=null;
	static boolean isParse=false;
	
	public SuperID_Process(Map<String, Set<Element> > superId2Question, Map<String, String> superId2FileName, boolean parse, AbstractSequenceClassifier<CoreLabel> namedEntityRcg, String superId) throws ClassCastException, ClassNotFoundException, IOException {
		if(superId2Q==null) superId2Q = superId2Question;
		if(superId2File==null) superId2File = superId2FileName;
		isParse = parse;
		ner = namedEntityRcg;
		sId=superId;
	}

	
	public void run(){
		//map passage id -> MCQuestion
		Map<String, Set<Element>> passage2Q = new HashMap<String, Set<Element>>();
		for(Element mcq : superId2Q.get(sId))for(Element doc : mcq.getElementsByTag(CONTEXT_ART_TAG)){
			String id = doc.attr(ID_TAG).toLowerCase(); 
			if(id.contains(sId)){
				if(!passage2Q.containsKey(id)) passage2Q.put(id, new HashSet<Element>() );
				passage2Q.get(id).add(mcq);
			}
		}
		
		//sort passage by id
		List<String> pIds = new ArrayList<String>(passage2Q.keySet());
		Collections.sort(pIds);
		
		//read file for each id
		try {
			InputStream fileStream = new FileInputStream(superId2File.get(sId));
			InputStream gzipStream = new GZIPInputStream(fileStream);
			Reader decoder = new InputStreamReader(gzipStream, "utf-8");
			BufferedReader br = new BufferedReader(decoder);
			for(String pId : pIds){
				String txt = findPassage(pId, br);
				
				//for unsorted file
				if(txt==null) txt = findPassage(pId, new File(superId2File.get(sId)));
				
				//put back txt into mcq
				for(Element mcq : passage2Q.get(pId))for(Element doc : mcq.getElementsByTag(CONTEXT_ART_TAG)){
					if(doc.attr(ID_TAG).equalsIgnoreCase(pId)){
						String parsed = new String(txt);
						if(isParse)try{
							parsed = parse(txt, mcq);
						}catch(ParseException e){
							e.printStackTrace();
						}
						doc.text(parsed.trim());
					}
				}
			}
			br.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
//	private Element parse(Element question, String[] cands) throws ParseException{
//		
//		//components of snt
//		List<HasWord> leftC = split(question.getElementsByTag(LEFT_CONTEXT_TAG).first().text());
//		List<HasWord> leftB = split(question.getElementsByTag(LEFT_BLANK_TAG).first().text());
//		List<HasWord> blank = split(question.getElementsByTag(BLANK_TAG).first().text());
//		List<HasWord> rightB= split(question.getElementsByTag(RIGHT_BLANK_TAG).first().text());
//		List<HasWord> rightC= split(question.getElementsByTag(RIGHT_CONTEXT_TAG).first().text());
//		
//		//marge them into snt
//		List<HasWord> snt = new ArrayList<HasWord>();
//		snt.addAll(leftC);
//		snt.addAll(leftB);
//		snt.addAll(blank);
//		snt.addAll(rightB);
//		snt.addAll(rightC);
//		
//		//NE recoginize
//		List<CoreLabel> neLabels = ner.classifySentence(snt);
//		
//		//tagging only on context
//		List<HasWord> leftTagged = tagging( snt.subList(0, leftC.size()), neLabels.subList(0, leftC.size()), cands );
//		question.getElementsByTag(LEFT_CONTEXT_TAG).first().text( toTxt(leftTagged) );
//		int mrgn=leftC.size()+leftB.size()+blank.size()+rightB.size();
//		List<HasWord> rightTagged = tagging( snt.subList(mrgn,snt.size()), neLabels.subList(mrgn, neLabels.size()), cands );
//		question.getElementsByTag(RIGHT_CONTEXT_TAG).first().text( toTxt(rightTagged) );
//		
//		return question;
//	}
	
	//return empty list if txt is not splitable
//	private List<HasWord> split(String txt){
//		List<HasWord> re =new ArrayList<HasWord>();
//		if(txt==null || txt.isEmpty() ) return re;
//		
//		String[] words = txt.split(" ");
//		for(String w: words) re.add(new Word(w));
//		return re;
//	}
	
	private String parse(String txt, Element mcq) throws ParseException{
		List<HasWord> parsed = new ArrayList<HasWord>();
		String[] cands = rawCands(mcq);
		
		DocumentPreprocessor dp = new DocumentPreprocessor(new StringReader(txt));
		for(Iterator<List<HasWord>> it = dp.iterator();it.hasNext();){
			List<HasWord> snt = it.next();
			parsed.addAll(parse(snt, cands));
		}
		
		return toTxt(parsed);
	}
	
	private String toTxt(List<HasWord> seq){
		StringBuffer sb = new StringBuffer();
		for(HasWord w : seq) sb.append(w.word().toString()+" ");
		return sb.toString().trim();
	}
	
	private String[] rawCands(Element mcq){
		//initialize slots for candidate answers
		String[] cands =  new String[mcq.getElementsByTag(CAND_TAG).size()];
		for(int i=0;i<cands.length;i++) cands[i]="";
	
		//fill it with index 
		for(Element e : mcq.getElementsByTag(CAND_TAG)) cands[Integer.parseInt(e.attr(INDEX_TAG))]=e.text();
		return cands;
	}
	
	private List<HasWord> parse(List<HasWord> snt , String[] rawCands) throws ParseException{
		//NE recoginize
		List<CoreLabel> neLabels = ner.classifySentence(snt);
		
		return tagging(snt, neLabels, rawCands);
	}
	
	/**
	 * return idx of candidate answer which corresponds to the phrase
	 *  -1 if there is no such candidate answer
	 * @param ph
	 * @param cands
	 * @return
	 */
	private int isCandAns(List<HasWord> ph, String[] rawCands){
		for(int i=0;i<rawCands.length;i++) if(isSubset(ph, rawCands[i])){ //if cand[j] has all words in the phrase
			return i;
		}
		return -1;
	}
	
	private boolean isSubset(List<HasWord> smaller, String larger){
		for(HasWord w: smaller){
			if(!larger.toLowerCase().contains(w.word().toString().toLowerCase())){
				return false;
			}
		}
		return true;
	}
	
	private List<HasWord> tagging(List<HasWord> snt, List<CoreLabel> neLabels, String[] rawCands) throws ParseException{
		if(neLabels.size()!=snt.size()) throw new ParseException("invalid NE recoginition");
		
		List<HasWord> anonySnt = new ArrayList<HasWord>();
		
		for(int i=0;i<snt.size();i++){
			String type = neLabels.get(i).get(CoreAnnotations.AnswerAnnotation.class);
			
			//NE label is not target
			if(!type.equalsIgnoreCase(ANS_TYPE)){
				anonySnt.add(snt.get(i));
			}else{
				
				//get NE
				List<HasWord> ph  =new ArrayList<HasWord>();
				for(int j=0;type!=null && i+j<snt.size() && type.equalsIgnoreCase(neLabels.get(i+j).get(CoreAnnotations.AnswerAnnotation.class));j++){
					ph.add(snt.get(i+j));
				}
											
				//check if NE is in cand ans
				int id=isCandAns(ph, rawCands);
				if(id!=-1){
					anonySnt.add(new Word(ENTITY_TAG_HEAD+id));
					anonySnt.addAll(ph);
					anonySnt.add(new Word(id+ENTITY_TAG_TAIL));
					
					//anonySnt.add(new Word(ENTITY_TAG+id+":"+connect(ph)));
				}else{
					anonySnt.addAll(ph);
				}
				
				i=i+ph.size()-1;
			}
		}
		
		return anonySnt;
	}
	
//	private static String connect(List<HasWord> ph){
//		StringBuffer sb =new StringBuffer();
//		for(int i=0;i<ph.size()-1;i++) sb.append(connectForm(ph.get(i).word())+"_");
//		sb.append(connectForm(ph.get(ph.size()-1).word()));
//		return sb.toString();
//	}
	
//	private static String connectForm(String word){
//		return word.replaceAll(" +","_");
//	}
	
	private String findPassage(String id, BufferedReader br) throws IOException {
		StringBuffer txt = new StringBuffer();
		//ancar tags
		String ancor = "<DOC id=\""+id.toUpperCase()+"\"";
		
		String line = br.readLine();
		while(line!=null){
			//Find id
			if(line.contains(ancor)){
				//find text
				while(line!=null && !line.contains(TEXT) && !line.contains(ID_TAIL)){
					line=br.readLine();
				}
				
				//store text
				if(line.contains(TEXT)){
					txt.append(line.replace(TEXT, " "));
					line = br.readLine();
				}else{
					break;
				}
				while(!line.contains(TEXT_TAIL)){
					line =line.replace(CORRUPTION, " ");
					line =line.replace(CORRUPTION_TAIL, " ");
					line =line.replace(PARAGRAPH, " ");
					line =line.replace(PARAGRAPH_TAIL, " ");
					line = line.trim();
					
					if(!line.isEmpty()) txt.append(line.trim()+" ");
					line = br.readLine();
				}
				txt.append(line.replace(TEXT_TAIL, " "));
				break;
			}
			line = br.readLine();
		}
		
		//if(line==null) throw new Exception("passage id not found : "+id);
		if(line==null) return null;
		
		String re = txt.toString();
		re=re.replaceAll("\t", " ");
		re=re.trim().replaceAll(" +", " "); 
		return unescape(re);
	}
	
	private String findPassage(String id, File xml) throws Exception{
		//Read xml file
		InputStream fileStream = new FileInputStream(xml);
		InputStream gzipStream = new GZIPInputStream(fileStream);
		Reader decoder = new InputStreamReader(gzipStream, "utf-8");
		BufferedReader br = new BufferedReader(decoder);
		
		String txt = findPassage(id, br);
		br.close();
			
		return txt;
	}

	/**
	 * this unexcape the followings,
	 * 	「&」----&amp
	 *	「>」----&gt;
	 *	「<」----&lt;
	 * @param txt
	 * @return
	 */
	private String unescape(String txt){
		String out = txt;
		out = out.replace("&amp;", "&");
		out = out.replace("&gt;", ">");
		out = out.replace("&lt;", "<");
		out = out.replace("&AMP;", "&");
		out = out.replace("&GT;", ">");
		out = out.replace("&LT;", "<");
		out = out.replace("&#8264;", "⁈");
		return out;
	}
}
