#!/bin/bash
#================================================
#	this is script to re-generate 
#	Who Did What (WDW) dataset using 
#	Gigaword
#
#		1st arg : path/to/gigaword
#================================================

if [ ! -d $1 ]; then
	echo $1" is not Gigaword directory"
	exit
fi
gigaDir=$1
n_thread=4

echo "Gigaword dir : "$gigaDir

#check base dir
root=`dirname $0`
data="/who_did_what_targeted"
if [ -d $root$data ]; then
	echo $root$data "already exists"
	exit
fi
mkdir $root$data

#mkdir for each data set type
keys="/keys"
for dtype in `ls $root$keys` ; do
	mkdir $root$data"/"$dtype
done

#compile java
libs=""
for l in `ls lib/*.jar`; do
        libs="${l}:${libs}"
done
libs="${libs}src"
javac -cp "${libs}" src/FindPassages.java

#key 2 file
for dtype in `ls $root/keys` ; do
	for d in $root$keys/$dtype/*_key.xml ; do
		k=$d
		d=${d##*/}
		t=$root$data"/"$dtype"/"${d%%_*}".xml"
		echo $k"->"$t
		java -Xmx5g -cp "${libs}" FindPassages -keyFile $k -gigaDir $gigaDir -outFile $t -ner `ls lib/*.crf.ser.gz` -parse
	done
done

#check sum

for dtype in `ls $root$keys` ; do
	for d in $root$keys/$dtype/*.xml.targeted_hk ; do
		k=$d
		d=${d##*/}
		t=$root$data"/"$dtype"/"${d%%.targeted_hk}
		echo "check : $t with $k"
		md5=`md5sum $t`
		md5=${md5%% *}
		if [ `cat $k` = $md5 ];	then
			echo "OK"
		else
			echo "something different"
		fi
	done
done
echo "The dataset is extracted into :"$root$data
